from datetime import timedelta
from math import acos, cos, degrees, radians, sin

import ephem

from .satellite import earth_radius_at_latitude
from .utils.curses import bresenham, closest_color, fill_outline
from .utils.geometry import (
    cartesian_to_latlon,
    latlon_to_cartesian,
    latlon_to_spherical,
    spherical_to_cartesian,
)
from .utils.text import format_seconds


GRID_LATITUDES = (60, 30, 0, -30, -60)
GRID_LONGITUDES = (150, 120, 90, 60, 30, 0, -30, -60, -90, -120, -150)


class InfoPanel(list):
    pass


def draw_coverage(layer, body, satellite, time, steps=100):
    interval = satellite.orbital_period.total_seconds() / steps
    footprints = set([])

    for i in range(steps):
        satellite.compute(time, plus_seconds=interval * i)
        earth_radius = earth_radius_at_latitude(satellite.latitude)
        horizon_radius = acos(earth_radius / (earth_radius + satellite.altitude))
        sat_xy = body.from_latlon(satellite.latitude, satellite.longitude)
        sat_footprints = []

        for hx, hy, hz in cartesian_rotation(
            satellite.latitude,
            satellite.longitude,
            horizon_radius,
            steps=int(body.width / 4),  # somewhat arbitrary
        ):
            sat_footprints.append(body.from_latlon(*cartesian_to_latlon(hx, hy, hz)))

        sat_outline = set(bresenham(sat_footprints, body.width, body.height, connect_ends=True))
        fill_outline(sat_xy, sat_outline, body.width, body.height)
        footprints.update(sat_outline)

    # now we have gathered the swept area, but we want to hide areas
    # *not* covered during the orbit
    inverted_footprints = []
    for x in range(body.width):
        for y in range(body.height):
            if (x, y) not in footprints:
                inverted_footprints.append((x, y))

    for x, y in inverted_footprints:
        layer.draw(x, y, "•", 94)

    # reset satellite to current position
    satellite.compute(time)


def draw_grid(layer, body):
    latitudes = []
    longitudes = []
    for lat in GRID_LATITUDES:
        x, y = body.from_latlon(lat, 0)
        latitudes.append(y)
    for lon in GRID_LONGITUDES:
        x, y = body.from_latlon(0, lon)
        longitudes.append(x)

    for x in range(body.width-1):
        for y in range(body.height):
            if body.map[x][y][3] is None:
                if x in longitudes and y in latitudes:
                    layer.draw(x, y, "+", 234)
                elif x in longitudes:
                    layer.draw(x, y, "|", 234)
                elif y in latitudes:
                    layer.draw(x, y, "─", 234)


def draw_info(
    layer,
    body,
    time,
    observer_latitude=None,
    observer_longitude=None,
    satellite=None,
):
    if satellite is not None:
        text_basic = InfoPanel()
        text_basic.append(satellite.name)
        text_basic.append("---")
        text_basic.append("Latitude:")
        text_basic.append("  {:.6f}°".format(satellite.latitude))
        text_basic.append("")
        text_basic.append("Longitude:")
        text_basic.append("  {:.6f}°".format(satellite.longitude))
        text_basic.append("")
        text_basic.append("Altitude:")
        text_basic.append("  {:,.2f}km".format(satellite.altitude / 1000))
        text_basic.append("")
        text_basic.append("Velocity:")
        text_basic.append("  {:,.2f}m/s".format(satellite.velocity))
        text_basic.append("")
        text_basic.append("Orbital period:")
        text_basic.append("  " + format_seconds(satellite.orbital_period.total_seconds()))
        text_basic.top = True
        text_basic.left = True

        text_apsides = InfoPanel()
        text_apsides.append("Apogee altitude:")
        text_apsides.append("  {:,.2f}km".format(satellite.apoapsis_altitude / 1000))
        text_apsides.append("")
        text_apsides.append("Perigee altitude:")
        text_apsides.append("  {:,.2f}km".format(satellite.periapsis_altitude / 1000))
        text_apsides.append("")
        text_apsides.append("Time to perigee:")
        text_apsides.append("  " + format_seconds(satellite.time_to_periapsis.total_seconds()))
        text_apsides.append("")
        text_apsides.append("Time since perigee:")
        text_apsides.append("  " + format_seconds(satellite.time_since_periapsis.total_seconds()))
        text_apsides.append("")
        text_apsides.append("Time to apogee:")
        text_apsides.append("  " + format_seconds(satellite.time_to_apoapsis.total_seconds()))
        text_apsides.append("")
        text_apsides.append("Time since apogee:")
        text_apsides.append("  " + format_seconds(satellite.time_since_apoapsis.total_seconds()))
        text_apsides.top = False
        text_apsides.left = False

        text_params = InfoPanel()
        text_params.append("Inclination:")
        text_params.append("  {:.4f}°".format(degrees(satellite.inclination)))
        text_params.append("")
        text_params.append("RA of asc node:")
        text_params.append("  {:.4f}°".format(
            degrees(satellite.right_ascension_of_ascending_node)
        ))
        text_params.append("")
        text_params.append("Arg of periapsis:")
        text_params.append("  {:.4f}°".format(degrees(satellite.argument_of_periapsis)))
        text_params.append("")
        text_params.append("Eccentricity:")
        text_params.append("  {:.7f}".format(satellite.eccentricity))
        text_params.append("")
        text_params.append("Semi-major axis:")
        text_params.append("  {:,.2f}km".format(satellite.semi_major_axis / 1000))
        text_params.append("")
        text_params.append("Mean anomaly @epoch:")
        text_params.append("  {:.4f}°".format(degrees(satellite.mean_anomaly_at_epoch)))
        text_params.append("")
        text_params.append("Epoch (UTC):")
        text_params.append(satellite.epoch.strftime("  %Y-%m-%d %H:%M:%S"))
        text_params.append("")
        text_params.append("Current time (UTC):")
        text_params.append(time.strftime("  %Y-%m-%d %H:%M:%S"))
        text_params.top = False
        text_params.left = True

        panels = [text_params, text_apsides, text_basic]

    else:  # no satellite
        text_time = InfoPanel()
        text_time.append("Time (UTC):")
        text_time.append(time.strftime("  %Y-%m-%d %H:%M:%S"))
        text_time.top = False
        text_time.left = True

        panels = [text_time]

    if observer_latitude is not None and observer_longitude is not None:
        text_observer = InfoPanel()
        text_observer.append("Observer")
        text_observer.append("---")
        text_observer.append("Latitude:")
        text_observer.append("  {:.6f}°".format(observer_latitude))
        text_observer.append("")
        text_observer.append("Longitude:")
        text_observer.append("  {:.6f}°".format(observer_longitude))
        if satellite is not None:
            text_observer.append("")
            text_observer.append("Azimuth:")
            text_observer.append("  {:.2f}°".format(degrees(satellite.observer_azimuth)))
            text_observer.append("")
            text_observer.append("Altitude:")
            text_observer.append("  {:.2f}°".format(degrees(satellite.observer_altitude)))
            if satellite.acquisition_of_signal is not None and \
                    satellite.loss_of_signal is not None:
                text_observer.append("")
                text_observer.append("AOS:")
                text_observer.append("  {}".format(
                    format_seconds((
                        satellite.acquisition_of_signal.datetime() - time
                    ).total_seconds())
                ))
                text_observer.append("")
                text_observer.append("LOS:")
                text_observer.append("  {}".format(
                    format_seconds((
                        satellite.loss_of_signal.datetime() - time
                    ).total_seconds())
                ))
        text_observer.top = True
        text_observer.left = False
        panels.append(text_observer)

    for text in panels:
        longest_line = max(map(len, text))

        text.padded_lines = []
        for line in text:
            if line == "---":
                text.padded_lines.append("┣" + "─" * (longest_line+2) + "┫")
            else:
                text.padded_lines.append("┃ " + line.ljust(longest_line+1) + "┃")
        text.padded_lines.insert(0, "┏" + "─" * (longest_line+2) + "┓")
        text.padded_lines.append("┗" + "─" * (longest_line+2) + "┛")

        if text.left:
            text.x = 2
        else:
            text.x = body.width - len(text.padded_lines[0]) - 2
        if text.top:
            text.y = 1
        else:
            text.y = body.height - len(text.padded_lines) - 1

        for y, line in enumerate(text.padded_lines):
            for x, char in enumerate(line):
                layer.draw(text.x + x, text.y + y, char, 0)


def draw_map(layer, body, time, night=True, topo=True):
    for x in range(body.width):
        for y in range(body.height):
            if night:
                sun = ephem.Sun()
                obs = ephem.Observer()
                obs.pressure = 0
                lat, lon = body.to_latlon(x, y)
                obs.lat = "{:.8f}".format(lat)
                obs.lon = "{:.8f}".format(lon)
                obs.date = time
                sun.compute(obs)
                # astronomical twilight starts at -18°
                # -0.3141592 = radians(-18)
                night_factor = max(min(sun.alt, 0), -0.3141592) / -0.3141592
            else:
                night_factor = -1

            if body.map[x][y][3] is not None:
                if topo is True:
                    r, g, b, color = body.map[x][y][:4]
                else:
                    r, g, b, color = 0, 249, 114, 48
                if night_factor > -0.001:
                    night_r = 0
                    night_g = g * 0.2
                    night_b = min(b + 40, 255)
                    effective_r = ((1 - night_factor) * r) + (night_factor * night_r)
                    effective_g = ((1 - night_factor) * g) + (night_factor * night_g)
                    effective_b = ((1 - night_factor) * b) + (night_factor * night_b)
                    color = closest_color(effective_r, effective_g, effective_b)
                layer.draw(x, y, "•", color)


def draw_orbits(
    layer,
    body,
    satellite,
    time,
    orbit_ascdesc=False,
    orbits=0,
    orbit_resolution="/70",
):
    if orbits == 0:
        return
    orbit_offset = timedelta()
    continuous = False
    if orbit_resolution.endswith("+"):
        continuous = True
        orbit_resolution = orbit_resolution.rstrip("+")
    if orbit_resolution.startswith("/"):
        orbit_increment = satellite.orbital_period / int(orbit_resolution[1:])
    else:
        orbit_increment = timedelta(minutes=float(orbit_resolution))

    satellite.compute(time)
    previous_altitude = satellite.altitude

    orbit_markers = []

    while orbit_offset < satellite.orbital_period * orbits + orbit_increment:
        satellite.compute(time, plus_seconds=orbit_offset.total_seconds())

        if orbit_ascdesc:
            if satellite.altitude - previous_altitude >= 0:
                char = "+"
            else:
                char = "-"
            previous_altitude = satellite.altitude
        else:
            char = "•"

        orbit_markers.append((body.from_latlon(satellite.latitude, satellite.longitude), char))
        orbit_offset += orbit_increment

    if continuous:
        orbit_marker_dict = dict(orbit_markers)
        orbit_markers = bresenham(
            [point for point, char in orbit_markers],
            body.width,
            body.height,
        )
        char = "•"
        for x, y in orbit_markers:
            try:
                char = orbit_marker_dict[(x, y)]
                interpolated = False
            except KeyError:
                interpolated = True
            color = 131 if interpolated else 209
            layer.draw(x, y, char, color)
    else:
        for point, char in orbit_markers:
            layer.draw(point[0], point[1], char, 209)

    # reset values to current
    satellite.compute(time)


def draw_apsides(layer, body, satellite):
    try:
        x, y = body.from_latlon(satellite.apoapsis_latitude, satellite.apoapsis_longitude)
        layer.draw(x, y, "A", 167)
    except ValueError:
        pass
    try:
        x, y = body.from_latlon(satellite.periapsis_latitude, satellite.periapsis_longitude)
        layer.draw(x, y, "P", 167)
    except ValueError:
        pass


def draw_footprint(layer, body, satellite):
    earth_radius = earth_radius_at_latitude(satellite.latitude)
    horizon_radius = acos(earth_radius / (earth_radius + satellite.altitude))
    footprint_markers = []

    for hx, hy, hz in cartesian_rotation(
        satellite.latitude,
        satellite.longitude,
        horizon_radius,
        steps=int(body.width / 4),  # somewhat arbitrary
    ):
        footprint_markers.append(body.from_latlon(*cartesian_to_latlon(hx, hy, hz)))

    footprint_markers = bresenham(footprint_markers, body.width, body.height, connect_ends=True)

    for x, y in footprint_markers:
        layer.draw(x, y, "•", 239)


def cartesian_rotation(lat, lon, r, steps=128):
    """
    Internally converts to Cartesian coordinates and applies a rotation
    matrix to yield a number of points (equal to steps) on the small
    circle described by the given latlon and radius (the latter being an
    angle as well).

    Math from github.com/vain/asciiworld.
    """
    # Get latitude of one point on the small circle. We can easily do
    # this by adjusting the latitude but we have to avoid pushing over a
    # pole.
    if lat > 0:
        slat = lat - degrees(r)
    else:
        slat = lat + degrees(r)

    # Geographic coordinates to spherical coordinates.
    s_theta, s_phi = latlon_to_spherical(lat, lon)

    # Cartesian coordinates of rotation axis.
    rx, ry, rz = spherical_to_cartesian(s_theta, s_phi)

    # Rotation matrix around r{x,y,z} by alpha.
    alpha = radians(360 / steps)

    M = []
    M.append(rx**2 * (1 - cos(alpha)) + cos(alpha))
    M.append(ry * rx * (1 - cos(alpha)) + rz * sin(alpha))
    M.append(rz * rx * (1 - cos(alpha)) - ry * sin(alpha))

    M.append(rx * ry * (1 - cos(alpha)) - rz * sin(alpha))
    M.append(ry**2 * (1 - cos(alpha)) + cos(alpha))
    M.append(rz * ry * (1 - cos(alpha)) + rx * sin(alpha))

    M.append(rx * rz * (1 - cos(alpha)) + ry * sin(alpha))
    M.append(ry * rz * (1 - cos(alpha)) - rx * sin(alpha))
    M.append(rz**2 * (1 - cos(alpha)) + cos(alpha))

    # Cartesian coordinates of initial vector.
    px, py, pz = latlon_to_cartesian(slat, lon)

    for i in range(steps):
        # Rotate p{x,y,z}.
        p2x = px * M[0] + py * M[3] + pz * M[6]
        p2y = px * M[1] + py * M[4] + pz * M[7]
        p2z = px * M[2] + py * M[5] + pz * M[8]

        # For acos(), force p2z back into [-1, 1] which *might* happen
        # due to precision errors.
        p2z_fixed = max(-1, min(1, p2z))

        # Use rotated p{x,y,z} as basis for next rotation.
        px = p2x
        py = p2y
        pz = p2z

        yield p2x, p2y, p2z_fixed


def draw_satellite(layer, body, satellite):
    try:
        x, y = body.from_latlon(satellite.latitude, satellite.longitude)
        layer.draw(x, y, "X", 16)
    except ValueError:
        pass


def draw_crosshair(layer, body, satellite):
    try:
        x, y = body.from_latlon(satellite.latitude, satellite.longitude)
    except ValueError:
        return
    for i in range(body.width-1):
        if body.map[i][y][3] is None:
            layer.draw(i, y, "─", 235)
    for i in range(body.height):
        if body.map[x][i][3] is None:
            layer.draw(x, i, "|", 235)


def draw_location(layer, body, lat, lon):
    if lat and lon:
        x, y = body.from_latlon(lat, lon)
        layer.draw(x, y, "•", 2)



from datetime import datetime, timedelta
from math import asin, atan2, cos, degrees, pi, radians, sin, sqrt
import re

import ephem
from requests import get


ALIASES = {
    'hst': 20580,
    'hubble': 20580,
    'iss': 25544,
    'landsat7': 25682,
    'landsat8': 39084,
    'qzs-1': 37158,
    'qzs1': 37158,
    'smap': 40376,
    'tiangong': 37820,
    'tiangong-1': 37820,
    'tiangong1': 37820,
}
EARTH_FLATTENING_COEFFICIENT = 0.003352891869237217
EARTH_RADIUS = 6378135
EARTH_SGP = 3.986004418e+14  # Standard gravitational parameter
KEPLER_ACCURACY = 1e-6
TLE_REGEX = re.compile("PRE>(.*)</PRE", flags=re.DOTALL)


def earth_radius_at_latitude(latitude):
    latitude = radians(abs(latitude))
    return EARTH_RADIUS * sqrt(
        1 -
        (2 * EARTH_FLATTENING_COEFFICIENT -
            (EARTH_FLATTENING_COEFFICIENT ** 2)
        )
        * (sin(latitude) ** 2)
    )


def epoch(yy, days):
    if int(yy) > 56:
        year = int("19" + yy)
    else:
        year = int("20" + yy)
    return datetime(year, 1, 1) + timedelta(days=float(days)-1)


def keplers_equation(mean_anomaly, eccentricity):
    eccentric_anomaly = mean_anomaly
    while True:
        d = eccentric_anomaly - eccentricity * sin(eccentric_anomaly) - mean_anomaly
        eccentric_anomaly -= d / (1 - eccentricity * cos(eccentric_anomaly))
        if abs(d) < KEPLER_ACCURACY:
            break
    return eccentric_anomaly


def semi_major_axis(mean_motion):
    return (EARTH_SGP / (mean_motion ** 2)) ** (1/3)


def orbital_velocity(semi_major_axis, altitude, latitude):
    return sqrt(
        EARTH_SGP * (
            (2 / (earth_radius_at_latitude(latitude) + altitude)) - (1 / semi_major_axis)
        )
    )


class EarthSatellite(object):
    def __init__(
        self,
        number,
        time,
        observer_latitude=0,
        observer_longitude=0,
        observer_elevation=0,
        tle_file=None,
    ):
        if tle_file is not None:
            with open(tle_file) as f:
                tle = f.read().strip().split("\n")
        elif number is not None:
            number = ALIASES.get(number, number)
            raw_html = get("http://www.celestrak.com/cgi-bin/TLE.pl?CATNR={}".format(number)).text
            tle = TLE_REGEX.search(raw_html).group(1).strip().split("\n")
        else:
            raise ValueError("No SATCAT number or TLE file provided")
        self._satellite = ephem.readtle(*tle)
        self.argument_of_periapsis = float(self._satellite._ap.norm)
        self.eccentricity = self._satellite._e
        self.epoch = self._satellite._epoch.datetime()
        self.inclination = float(self._satellite._inc.norm)
        self.mean_anomaly_at_epoch = float(self._satellite._M.norm)
        self.mean_motion_revs_per_day = self._satellite._n
        self.name = tle[0].strip()
        self.observer_elevation = observer_elevation
        self.observer_latitude = observer_latitude
        self.observer_longitude = observer_longitude
        self.orbital_period = timedelta(days=1) / self.mean_motion_revs_per_day
        self.mean_motion = 2 * pi / self.orbital_period.total_seconds()
        self.apoapsis_latitude = degrees(asin(
            sin(self.argument_of_periapsis + pi) * sin(self.inclination)
        ))
        self.periapsis_latitude = degrees(asin(
            sin(self.argument_of_periapsis) * sin(self.inclination)
        ))
        self.right_ascension_of_ascending_node = float(self._satellite._raan.norm)
        self.semi_major_axis = semi_major_axis(self.mean_motion)
        self.apoapsis_altitude = self.semi_major_axis * (1 + self.eccentricity) - \
                                 earth_radius_at_latitude(self.apoapsis_latitude)
        self.periapsis_altitude = self.semi_major_axis * (1 - self.eccentricity) - \
                                  earth_radius_at_latitude(self.periapsis_latitude)
        self.compute(time)

    def compute(self, time, plus_seconds=0):
        target_time = time + timedelta(seconds=plus_seconds)
        self._satellite.compute(target_time)
        self.altitude = self._satellite.elevation
        self.latitude = degrees(self._satellite.sublat)
        self.longitude = degrees(self._satellite.sublong)
        self.mean_anomaly = (
            self.mean_anomaly_at_epoch +
            self.mean_motion * (
                (target_time - self.epoch).total_seconds() %
                self.orbital_period.total_seconds()
            )
        ) % (2 * pi)
        self.eccentric_anomaly = keplers_equation(self.mean_anomaly, self.eccentricity)
        self.true_anomaly = 2 * atan2(
            sqrt(1 + self.eccentricity) * sin(self.eccentric_anomaly / 2),
            sqrt(1 - self.eccentricity) * cos(self.eccentric_anomaly / 2)
        )
        self.time_since_periapsis = timedelta(seconds=self.mean_anomaly / self.mean_motion)
        self.time_to_periapsis = timedelta(seconds=self.orbital_period.total_seconds() -
                                                   self.time_since_periapsis.total_seconds())
        self.time_since_apoapsis = timedelta(seconds=(self.mean_anomaly + pi) % (2 * pi) /
                                                     self.mean_motion)
        self.time_to_apoapsis = timedelta(seconds=self.orbital_period.total_seconds() -
                                                  self.time_since_apoapsis.total_seconds())
        self.velocity = orbital_velocity(self.semi_major_axis, self.altitude, self.latitude)

        self._satellite.compute(target_time + self.time_to_periapsis)
        self.periapsis_latitude = degrees(self._satellite.sublat)
        self.periapsis_longitude = degrees(self._satellite.sublong)

        self._satellite.compute(target_time + self.time_to_apoapsis)
        self.apoapsis_latitude = degrees(self._satellite.sublat)
        self.apoapsis_longitude = degrees(self._satellite.sublong)

        if (
            self.observer_latitude is not None and
            self.observer_longitude is not None
        ):
            observer = ephem.Observer()
            observer.date = target_time
            observer.elevation = self.observer_elevation
            observer.lat = str(self.observer_latitude)
            observer.lon = str(self.observer_longitude)
            self._satellite.compute(observer)
            self.observer_azimuth = float(self._satellite.az.norm)
            self.observer_altitude = float(self._satellite.alt.znorm)
            self.acquisition_of_signal = self._satellite.rise_time
            self.loss_of_signal = self._satellite.set_time



import curses
from datetime import datetime, timedelta
from queue import Empty
from threading import Thread

import click
from requests import get

from . import VERSION_STRING
from .body import BODY_MAP
from .draw import (
    draw_apsides,
    draw_coverage,
    draw_crosshair,
    draw_footprint,
    draw_grid,
    draw_info,
    draw_location,
    draw_map,
    draw_orbits,
    draw_satellite,
)
from .layer import Layer, pixel_from_layers
from .satellite import ALIASES, EarthSatellite
from .utils.curses import graceful_ctrlc, input_thread_body, setup
from .utils.curses import (
    INPUT_CYCLE_ORBITS,
    INPUT_EXIT,
    INPUT_TIME_MINUS_LONG,
    INPUT_TIME_MINUS_SHORT,
    INPUT_TIME_PAUSE,
    INPUT_TIME_PLUS_LONG,
    INPUT_TIME_PLUS_SHORT,
    INPUT_TIME_RESET,
    INPUT_TOGGLE_COVERAGE,
    INPUT_TOGGLE_CROSSHAIR,
    INPUT_TOGGLE_FOOTPRINT,
    INPUT_TOGGLE_GRID,
    INPUT_TOGGLE_INFO,
    INPUT_TOGGLE_NIGHT,
    INPUT_TOGGLE_ORBIT_APSIDES,
    INPUT_TOGGLE_ORBIT_ASCDESC,
    INPUT_TOGGLE_TOPO,
)
from .utils.text import format_seconds


def check_for_resize(stdscr, body):
    start = datetime.now()
    height, width = stdscr.getmaxyx()

    if body.height != height or body.width != width:
        body = body.__class__(width, height)
        progress_str = "0.0"
        start = datetime.now()
        for progress in body.prepare_map():
            if progress_str != "{:.2f}".format(progress):
                progress_str = "{:.2f}".format(progress)
                elapsed_time = datetime.now() - start
                eta = (elapsed_time / max(progress, 0.01)) * (100 - progress)
                stdscr.erase()
                stdscr.insstr(0, 0, "Rendering map (ETA {}, {}%)...".format(
                    format_seconds(eta.total_seconds()),
                    progress_str,
                ))
                stdscr.refresh()
        return body, True
    else:
        return body, False


def redraw(stdscr, body, layers):
    stdscr.erase()
    for x in range(body.width):
        for y in range(body.height):
            char, color = pixel_from_layers(x, y, layers)
            stdscr.insstr(y, x, char, curses.color_pair(color))


@graceful_ctrlc
def render(
        stdscr,
        apsides=False,
        body="earth",
        coverage=False,
        crosshair=False,
        footprint=False,
        fps=1,
        grid=False,
        info=False,
        me=False,
        night=False,
        observer=None,
        orbit_ascdesc=False,
        orbit_res="/70",
        orbits=0,
        paused=False,
        satellite=None,
        tle=None,
        topo=False,
        **kwargs
):
    curses_lock, input_queue, quit_event = setup(stdscr)
    input_thread = Thread(
        args=(stdscr, input_queue, quit_event, curses_lock),
        target=input_thread_body,
    )
    input_thread.start()
    try:
        body = BODY_MAP[body.lower()](1, 1)
        if body.NAME != "Earth":
            night = False
            topo = False
            me = False
            satellite = None

        observer_latitude = None
        observer_longitude = None
        if me and observer is None:
            location_data = get("http://ip-api.com/json").json()
            observer_latitude = location_data['lat']
            observer_longitude = location_data['lon']
        if observer is not None:
            obs_latlon = observer.split()
            observer_latitude = float(obs_latlon[0])
            observer_longitude = float(obs_latlon[1])

        time_offset = timedelta(0)
        time = datetime.utcnow() + time_offset
        force_redraw = False

        if paused is True:
            paused = datetime.utcnow()
            force_redraw = True

        if satellite is None and tle is None:
            satellite_obj = None
        else:
            satellite_obj = EarthSatellite(
                satellite,
                time,
                observer_latitude=observer_latitude,
                observer_longitude=observer_longitude,
                tle_file=tle,
            )

        apsides_layer = Layer(draw_apsides, update_timeout=8)
        apsides_layer.hidden = not apsides
        coverage_layer = Layer(draw_coverage, update_timeout=10)
        coverage_layer.hidden = not coverage
        crosshair_layer = Layer(draw_crosshair)
        crosshair_layer.hidden = not crosshair
        footprint_layer = Layer(draw_footprint)
        footprint_layer.hidden = not footprint
        grid_layer = Layer(draw_grid, update_timeout=None)
        grid_layer.hidden = not grid
        info_layer = Layer(draw_info)
        info_layer.hidden = not info
        map_layer = Layer(draw_map, update_timeout=None)
        observer_layer = Layer(draw_location, update_timeout=None)
        orbit_layer = Layer(draw_orbits)
        satellite_layer = Layer(draw_satellite)
        satellite_layer.hidden = satellite_obj is None

        layers = [
            info_layer,
            satellite_layer,
            apsides_layer,
            observer_layer,
            footprint_layer,
            orbit_layer,
            coverage_layer,
            map_layer,
            crosshair_layer,
            grid_layer,
        ]

        while True:
            with curses_lock:
                body, did_resize = check_for_resize(stdscr, body)
            if did_resize:
                for layer in layers:
                    layer.last_updated = None

            draw_start = datetime.now()
            if not paused:
                time = datetime.utcnow() + time_offset
            if force_redraw:
                for layer in layers:
                    layer.last_updated = None
            if not paused or force_redraw:
                grid_layer.update(body)
                info_layer.update(
                    body,
                    time,
                    observer_latitude=observer_latitude,
                    observer_longitude=observer_longitude,
                    satellite=satellite_obj,
                )
                map_layer.update(body, time, night=night, topo=topo)
                observer_layer.update(body, observer_latitude, observer_longitude)

                if satellite_obj is not None:
                    apsides_layer.update(body, satellite_obj)
                    coverage_layer.update(body, satellite_obj, time)
                    crosshair_layer.update(body, satellite_obj)
                    footprint_layer.update(body, satellite_obj)
                    orbit_layer.update(
                        body,
                        satellite_obj,
                        time,
                        orbits=orbits,
                        orbit_ascdesc=orbit_ascdesc,
                        orbit_resolution=orbit_res,
                    )
                    satellite_layer.update(body, satellite_obj)
                    satellite_obj.compute(time)

                with curses_lock:
                    redraw(stdscr, body, layers)

            draw_time = (datetime.now() - draw_start).total_seconds()

            # get input
            try:
                input_action = input_queue.get(True, max(0, 1/fps - draw_time))
                # we just received an input that probably modified how
                # our screen is supposed to look, ergo we need to redraw
                force_redraw = True
            except Empty:
                input_action = None
                force_redraw = False

            # react to input
            if input_action == INPUT_CYCLE_ORBITS:
                orbits += 1
                orbits = orbits % 4
            elif input_action == INPUT_EXIT:
                break
            elif input_action == INPUT_TIME_MINUS_SHORT:
                if satellite_obj is None:
                    time_offset -= timedelta(minutes=30)
                else:
                    time_offset -= satellite_obj.orbital_period / 20
            elif input_action == INPUT_TIME_MINUS_LONG:
                if satellite_obj is None:
                    time_offset -= timedelta(hours=6)
                else:
                    time_offset -= satellite_obj.orbital_period / 2
            elif input_action == INPUT_TIME_PAUSE:
                if paused:
                    time_offset -= datetime.utcnow() - paused
                    paused = False
                else:
                    paused = datetime.utcnow()
            elif input_action == INPUT_TIME_PLUS_SHORT:
                if satellite_obj is None:
                    time_offset += timedelta(minutes=30)
                else:
                    time_offset += satellite_obj.orbital_period / 20
            elif input_action == INPUT_TIME_PLUS_LONG:
                if satellite_obj is None:
                    time_offset += timedelta(hours=6)
                else:
                    time_offset += satellite_obj.orbital_period / 2
            elif input_action == INPUT_TIME_RESET:
                time_offset = timedelta(0)
                if paused:
                    paused = time = datetime.utcnow()
            elif input_action == INPUT_TOGGLE_COVERAGE:
                coverage_layer.hidden = not coverage_layer.hidden
                coverage_layer.last_updated = None
            elif input_action == INPUT_TOGGLE_CROSSHAIR:
                crosshair_layer.hidden = not crosshair_layer.hidden
            elif input_action == INPUT_TOGGLE_FOOTPRINT:
                footprint_layer.hidden = not footprint_layer.hidden
            elif input_action == INPUT_TOGGLE_GRID:
                grid_layer.hidden = not grid_layer.hidden
            elif input_action == INPUT_TOGGLE_INFO:
                info_layer.hidden = not info_layer.hidden
            elif input_action == INPUT_TOGGLE_NIGHT:
                night = not night
                map_layer.last_updated = None
                if night:
                    map_layer.update_timeout = timedelta(seconds=60)
                else:
                    map_layer.update_timeout = None
            elif input_action == INPUT_TOGGLE_ORBIT_APSIDES:
                apsides_layer.hidden = not apsides_layer.hidden
                apsides_layer.last_updated = None
            elif input_action == INPUT_TOGGLE_ORBIT_ASCDESC:
                orbit_ascdesc = not orbit_ascdesc
                orbit_layer.last_updated = None
            elif input_action == INPUT_TOGGLE_TOPO:
                topo = not topo
                map_layer.last_updated = None
    finally:
        quit_event.set()
        input_thread.join()


def print_aliases(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    for alias in sorted(ALIASES.keys()):
        click.echo("{}: {}".format(alias, ALIASES[alias]))
    ctx.exit()


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(VERSION_STRING)
    ctx.exit()


@click.command()
@click.option("--aliases", is_flag=True, callback=print_aliases,
              expose_value=False, is_eager=True,
              help="Show all satellite aliases and exit")
@click.option("--apsides", is_flag=True, default=False,
              help="Draw apoapsis and periapsis markers")
@click.option("-b", "--body", default="earth", metavar="BODY",
              help="Which celestial body to draw: Earth, Moon or Mars "
                   "(defaults to Earth)")
@click.option("-c", "--coverage", is_flag=True, default=False,
              help="Show next-orbit coverage overlay")
@click.option("-f", "--footprint", is_flag=True, default=False,
              help="Draw satellite footprint/horizon")
@click.option("--fps", default=1, metavar="N",
              help="Frames per second (defaults to 1)")
@click.option("-g", "--grid", is_flag=True, default=False,
              help="Draw latitude/longitude grid")
@click.option("-i", "--info", is_flag=True, default=False,
              help="Show info panels")
@click.option("-m", "--me", is_flag=True, default=False,
              help="Auto-detect your location as observer")
@click.option("-n", "--night", is_flag=True, default=False,
              help="Shade night side")
@click.option("-o", "--orbits", default=0, metavar="N",
              help="Draw this many orbits ahead of the satellite")
@click.option("--orbit-ascdesc", is_flag=True, default=False,
              help="Draw orbits with ascent/descent markers")
@click.option("-O", "--observer", default=None, metavar="'LAT LON'",
              help="Space-separated latitude and longitude of an "
                   "observer; overrides IP-geolocation")
@click.option("-p", "--paused", is_flag=True, default=False,
              help="Start paused")
@click.option("-r", "--orbit-res", default="/70", metavar="[/]N[+]",
              help="Set distance of orbit markers: 'N' means N minutes, "
                   "'/N' means 1/Nth of orbital period, append a plus "
                   "sign to interpolate in between markers (defaults to /70)")
@click.option("-t", "--topo", is_flag=True, default=False,
              help="Enable coloring of topographical features")
@click.option("--tle", default=None, metavar="FILE",
              help="read TLE data from FILE instead of downloading it "
                   "(SATELLITE will have no effect and can be omitted)")
@click.option("-x", "--crosshair", is_flag=True, default=False,
              help="Draw crosshair around satellite location")
@click.option("--version", is_flag=True, callback=print_version,
              expose_value=False, is_eager=True,
              help="Show version and exit")
@click.argument('satellite', required=False)
def main(**kwargs):
    """
    \b
    Shows a world map tracking SATELLITE. Valid values for SATELLITE are
    numbers from http://www.celestrak.com/NORAD/elements/master.asp (for
    your convenience, a number of aliases have been provided).
    \b
    Example satellite aliases (find more with --aliases):
    \thubble\t\tHubble Space Telescope
    \tiss\t\tInternational Space Station
    \ttiangong\tTiangong-1 (Chinese space station)
    \b
    Hotkeys:
    \ta\tToggle apsides markers
    \tc\tToggle next-orbit coverage overlay
    \td\tToggle ascent/descent markers
    \tf\tToggle footprint (satellite horizon)
    \tg\tToggle latitude/longitude grid
    \ti\tToggle info panels
    \tn\tToggle night shading
    \to\tCycle through drawing 0-3 next orbits
    \tp\tPause/resume
    \tq\tQuit
    \tr\tReset plotted time to current
    \tt\tToggle topography
    \tx\tToggle crosshair
    \tleft\tSmall step back in time
    \tright\tSmall step forward in time
    \tdown\tLarge step back in time
    \tup\tLarge step forward in time
    """
    curses.wrapper(render, **kwargs)



from datetime import datetime, timedelta


def pixel_from_layers(x, y, layers):
    for layer in layers:
        if layer.hidden:
            continue
        try:
            return layer.content[(x, y)]
        except KeyError:
            pass
    return " ", 1


class Layer(object):
    def __init__(self, update_callback, update_timeout=0):
        self.content = {}
        self.hidden = False
        self.last_updated = None
        self.update_callback = update_callback
        if update_timeout is None:
            self.update_timeout = None
        else:
            self.update_timeout = timedelta(seconds=update_timeout)

    def draw(self, x, y, char, color):
        self.content[(x, y)] = (char, color)

    def update(self, *args, **kwargs):
        if (
            self.hidden or
            (self.last_updated and (
                self.update_timeout is None or
                datetime.now() - self.last_updated < self.update_timeout
            ))
        ):
            return

        self.last_updated = datetime.now()
        self.content = {}
        self.update_callback(self, *args, **kwargs)



from copy import copy
from os.path import dirname, expanduser, join
import shelve

from PIL import Image
import shapefile

from .utils.curses import closest_color
from .utils.geometry import latlon_to_spherical, point_in_poly, spherical_to_cartesian


MAP_CACHE = "~/.termtrack_map_cache"


class Body(object):
    def __init__(self, width, height):
        self.height = height
        self.width = width
        self.pixel_percentage = 100 / (self.width * self.height)
        self._img = Image.open(join(dirname(__file__), "data", self.COLORMAP))
        if self.SHAPEFILE is not None:
            self._sf = shapefile.Reader(join(dirname(__file__), "data", self.SHAPEFILE))

    def from_latlon(self, lat, lon):
        xrel = (lon + 180) / 360
        yrel = (-lat + 90) / 180
        x = round((self.width - 1) * xrel)
        y = round((self.height - 1) * yrel)
        return min(x, self.width - 1), min(y, self.height - 1)

    def prepare_map(self):
        map_cache = shelve.open(expanduser(MAP_CACHE))
        try:
            map_cache_key = "{}_{}x{}".format(self.NAME, self.width, self.height)
            if map_cache_key in map_cache:
                self.map = map_cache[map_cache_key]
                raise StopIteration()
            progress = 0.0
            empty_line = [None for i in range(self.height)]
            self.map = [copy(empty_line) for i in range(self.width)]

            img = self._img.resize((self.width, self.height))
            pixels = img.load()

            for x in range(self.width):
                for y in range(self.height):
                    yield progress
                    color = r = g = b = None
                    lat, lon = self._to_latlon(x, y)
                    if self.SHAPEFILE is not None:
                        for shape in self._sf.iterShapes():
                            if (
                                # for performance reasons we quickly check the
                                # bounding box before trying the more expensive
                                # point_in_poly() call
                                lat > shape.bbox[1] and
                                lat < shape.bbox[3] and
                                lon > shape.bbox[0] and
                                lon < shape.bbox[2]
                            ) and point_in_poly(lon, lat, shape.points):
                                r, g, b = pixels[x, y]
                                color = closest_color(r, g, b)
                                break
                    else:
                        r, g, b = pixels[x, y]
                        color = closest_color(r, g, b)
                    spherical = latlon_to_spherical(lat, lon)
                    cartesian = spherical_to_cartesian(*spherical)
                    self.map[x][y] = (r, g, b, color, (lat, lon), spherical, cartesian)
                    progress += self.pixel_percentage
                    yield progress
            map_cache[map_cache_key] = self.map
        finally:
            map_cache.close()

    def to_cartesian(self, x, y):
        return self.map[x][y][6]

    def to_latlon(self, x, y):
        return self.map[x][y][4]

    def _to_latlon(self, x, y):
        xrel = x / (self.width - 1)
        yrel = y / (self.height - 1)
        return (
            90 - yrel * 180,
            xrel * 360 - 180,
        )

    def to_spherical(self, x, y):
        return self.map[x][y][5]


class Earth(Body):
    NAME = "Earth"
    COLORMAP = "earth.jpg"
    SHAPEFILE = "ne_110m_land.shp"


class Mars(Body):
    NAME = "Mars"
    COLORMAP = "mars.jpg"
    SHAPEFILE = None


class Moon(Body):
    NAME = "Moon"
    COLORMAP = "moon.jpg"
    SHAPEFILE = None


BODY_MAP = {
    'earth': Earth,
    'luna': Moon,
    'mars': Mars,
    'moon': Moon,
    'terra': Earth,
}



from math import acos, atan2, cos, degrees, radians, sin, sqrt


def cartesian_to_latlon(x, y, z):
    return spherical_to_latlon(*cartesian_to_spherical(x, y, z))


def cartesian_to_spherical(x, y, z):
    return acos(z), atan2(y, x)


def latlon_to_cartesian(lat, lon):
    return spherical_to_cartesian(*latlon_to_spherical(lat, lon))


def latlon_to_spherical(lat, lon):
    return -radians(lat) + radians(90), radians(lon)


def point_distance(point1, point2):
    return sqrt(abs((point2[0] - point1[0]) ** 2 - (point2[1] - point1[1]) ** 2))


def point_in_poly(x, y, poly):
    n = len(poly)
    inside = False
    p1x, p1y = poly[0]
    for i in range(n+1):
        p2x, p2y = poly[i % n]
        if y > min(p1y, p2y):
            if y <= max(p1y, p2y):
                if x <= max(p1x, p2x):
                    if p1y != p2y:
                        xints = (y-p1y) * (p2x-p1x) / (p2y-p1y) + p1x
                    if p1x == p2x or x <= xints:
                        inside = not inside
        p1x, p1y = p2x, p2y
    return inside


def spherical_to_cartesian(theta, phi):
    x = sin(theta) * cos(phi)
    y = sin(theta) * sin(phi)
    z = cos(theta)
    return x, y, z


def spherical_to_latlon(theta, phi):
    return degrees(radians(90) - theta), degrees(phi)



from setuptools import find_packages, setup

setup(
    name="termtrack",
    version="0.6.1",
    description="Track Earth-orbiting satellites from your terminal",
    author="Torsten Rehn",
    author_email="torsten@rehn.email",
    license="GPLv3",
    url="https://github.com/trehn/termtrack",
    keywords=["terminal", "track", "tracking", "satellite", "orbit", "iss"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console :: Curses",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3",
        "Topic :: Utilities",
    ],
    packages=find_packages(),
    package_data={
        "termtrack": [
            "data/earth.jpg",
            "data/mars.jpg",
            "data/moon.jpg",
            "data/ne_110m_land.dbf",
            "data/ne_110m_land.shp",
            "data/ne_110m_land.shx",
        ],
    },
    install_requires=[
        "click >= 2.0",
        "Pillow >= 2.7.0",
        "pyephem >= 3.7.5.0",
        "pyshp >= 1.2.1",
        "requests >= 2.0.0",
    ],
    entry_points={
        'console_scripts': [
            "termtrack=termtrack.cli:main",
        ],
    },
)



def format_seconds(seconds, hide_seconds=False):
    """
    Returns a human-readable string representation of the given amount
    of seconds.
    """
    if seconds < 1:
        return "0s"
    output = ""
    for period, period_seconds in (
        ('y', 31557600),
        ('d', 86400),
        ('h', 3600),
        ('m', 60),
        ('s', 1),
    ):
        if seconds >= period_seconds and not (hide_seconds and period == 's'):
            output += str(int(seconds / period_seconds))
            output += period
            output += " "
            seconds = seconds % period_seconds
    return output.strip()
